import unittest
import numpy as np
from simsopt._core.util import *


class IsboolTests(unittest.TestCase):
    def test_basic(self):
        self.assertTrue(isbool(True))
        self.assertTrue(isbool(False))
        self.assertFalse(isbool("howdy"))
        a = np.array([True, False])
        b = np.array([1, 2, 3])
        self.assertTrue(isbool(a[0]))
        self.assertFalse(isbool(b[0]))


class IsnumberTests(unittest.TestCase):
    def test_basic(self):
        # Try some arguments that ARE numbers:
        self.assertTrue(isnumber(5))
        self.assertTrue(isnumber(5.0))
        a = np.array([1])
        b = np.array([1.0])
        self.assertTrue(isnumber(a[0]))
        self.assertTrue(isnumber(b[0]))

        # Try some arguments that are NOT numbers:
        self.assertFalse(isnumber("foo"))
        self.assertFalse(isnumber(object))
        self.assertFalse(isnumber([1, 2, 3]))


class UniqueTests(unittest.TestCase):
    def test_basic(self):
        self.assertEqual(unique([]), [])
        self.assertEqual(unique([5]), [5])
        self.assertEqual(unique([5, 5]), [5])
        self.assertEqual(unique([1, -3, 7, 2]), [1, -3, 7, 2])
        self.assertEqual(unique([1, -3, 7, 2, 1, -3, 7]), [1, -3, 7, 2])


class FiniteDifferenceStepsTests(unittest.TestCase):
    def test_abs_step(self):
        """
        Test situations in which abs_step should dominate.
        """
        N = 5
        x = np.zeros(N)
        abs_step = 1.0e-5
        np.testing.assert_allclose(finite_difference_steps(x,
                                                           abs_step=abs_step,
                                                           rel_step=0),
                                   np.full(N, abs_step))

        x = np.array([0.22703548, -0.09694371, 0.36863702,
                      -0.46309711, 0.164472, -0.08652083])
        np.testing.assert_allclose(finite_difference_steps(x,
                                                           abs_step=abs_step,
                                                           rel_step=0),
                                   np.full(6, abs_step))

    def test_rel_step(self):
        """
        Test situations in which rel_step should dominate.
        """
        x = np.array([0.1165368, -0.92130271, 1.09012617, 0.71410455,
                      -0.77997291, -0.69077027, -1.51686311])

        rel_step = 1.0e-5
        np.testing.assert_allclose(finite_difference_steps(x,
                                                           abs_step=0,
                                                           rel_step=rel_step),
                                   np.abs(x) * rel_step)

    def test_mixed(self):
        """
        Test situations in which rel_step should dominate for some elements
        while abs_step should dominate for others.
        """
        x = np.array([0.1165368, -0.92130271, 1.09012617, 0.71410455,
                      -0.77997291, -0.69077027, -1.51686311])

        x0 = np.array([0.08, 0.092130271, 0.109012617, 0.08,
                       0.08, 0.08, 0.151686311])

        np.testing.assert_allclose(finite_difference_steps(x,
                                                           abs_step=0.08,
                                                           rel_step=0.1),
                                   x0)

        x = np.array([0, 1.21923073, 0.85723015, -0.08222701,
                      0, -0.9645991, 0, 0])

        x0 = np.array([1e-7, 1.21923073e-3, 0.85723015e-3, 0.08222701e-3,
                       1e-7, 0.9645991e-3, 1e-7, 1e-7])

        np.testing.assert_allclose(finite_difference_steps(x,
                                                           abs_step=1.0e-7,
                                                           rel_step=1.0e-3),
                                   x0)

    def test_zero_step(self):
        """
        Make sure we handle the case of a step of size 0 arising when
        abs_step is 0 and an element of x is 0.
        """
        x = np.array([0.22703548, -0.09694371, 0.36863702,
                      -0.46309711, 0, -0.08652083])

        with self.assertRaises(ValueError):
            finite_difference_steps(x, abs_step=0, rel_step=1e-3)
        with self.assertRaises(ValueError):
            finite_difference_steps(x, abs_step=0, rel_step=0)

        x = np.zeros(1)
        with self.assertRaises(ValueError):
            finite_difference_steps(x, abs_step=0, rel_step=1e-3)
        with self.assertRaises(ValueError):
            finite_difference_steps(x, abs_step=0, rel_step=0)


class NestedListsToArrayTests(unittest.TestCase):
    def test_nested_lists_to_array(self):
        """
        Test the utility function used to convert 2D data in a fortran
        namelist extracted by f90nml to a 2D numpy array.
        """
        list_of_lists = [[42]]
        arr1 = nested_lists_to_array(list_of_lists)
        arr2 = np.array([[42]])
        np.testing.assert_allclose(arr1, arr2)

        list_of_lists = [[42], [1, 2, 3]]
        arr1 = nested_lists_to_array(list_of_lists)
        arr2 = np.array([[42, 0, 0],
                         [1, 2, 3]])
        np.testing.assert_allclose(arr1, arr2)

        list_of_lists = [[None, 42], [1, 2, 3]]
        arr1 = nested_lists_to_array(list_of_lists)
        arr2 = np.array([[0, 42, 0],
                         [1, 2, 3]])
        np.testing.assert_allclose(arr1, arr2)

        list_of_lists = [[42, 43, 44], [1, 2, 3]]
        arr1 = nested_lists_to_array(list_of_lists)
        arr2 = np.array([[42, 43, 44],
                         [1, 2, 3]])
        np.testing.assert_allclose(arr1, arr2)

        list_of_lists = [[42, 43, 44, 45], [1, 2, 3]]
        arr1 = nested_lists_to_array(list_of_lists)
        arr2 = np.array([[42, 43, 44, 45],
                         [1, 2, 3, 0]])
        np.testing.assert_allclose(arr1, arr2)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3

import logging
from simsopt.objectives.functions import Identity
from simsopt.objectives.least_squares import LeastSquaresProblem
from simsopt.solve.serial import least_squares_serial_solve

"""                                                                           
Minimize f(x,y,z) = ((x-1)/1)^2 + ((y-2)/2)^2 + ((z-3)/3)^2.                  
The optimum is at (x,y,z)=(1,2,3), and f=0 at this point.                     
"""

# To print out diagnostic information along the way, uncomment this
# next line:
#logging.basicConfig(level=logging.INFO)

# Define some Target objects that depend on Parameter objects. In the
# future these functions would involve codes like VMEC, but for now we
# just use the functions f(x) = x.
iden1 = Identity()
iden2 = Identity()
iden3 = Identity()

# Parameters are all not fixed by default, meaning they will not be
# optimized.  You can choose to exclude any subset of the parameters
# from the space of independent variables by setting their 'fixed'
# property to True.
#iden1.fixed[0] = True
#iden2.fixed[0] = True
#iden3.fixed[0] = True

# Each Target is then equipped with a shift and weight, to become a
# term in a least-squares objective function
term1 = (iden1, 1, 1)
term2 = (iden2, 2, 2)
term3 = (iden3, 3, 3)

# A list of terms are combined to form a nonlinear-least-squares problem.
prob = LeastSquaresProblem([term1, term2, term3])

# Solve the minimization problem:
least_squares_serial_solve(prob)

print("An optimum was found at x=", iden1.x, ", y=", iden2.x, \
      ", z=", iden3.x)
print("The minimum value of the objective function is ", prob.objective())



#!/usr/bin/env python3
from simsopt.field.biotsavart import BiotSavart
from simsopt.field.coil import coils_via_symmetries
from simsopt.geo.surfacerzfourier import SurfaceRZFourier
from simsopt.geo.qfmsurface import QfmSurface
from simsopt.geo.surfaceobjectives import QfmResidual, ToroidalFlux, Area, Volume
from simsopt.util.zoo import get_ncsx_data
import numpy as np
import os

"""
This example demonstrate how to compute a quadratic flux minimizing surfaces
from a magnetic field induced by coils. We start with an initial guess that
is just a tube around the magnetic axis using the NCSX coils. We first reduce
the penalty objective function to target a given toroidal flux using LBFGS.
The equality constrained objective is then reduced using SLSQP. This is repeated
for fixing the area and toroidal flux.
"""

curves, currents, ma = get_ncsx_data()
coils = coils_via_symmetries(curves, currents, 3, True)
bs = BiotSavart(coils)
bs_tf = BiotSavart(coils)

mpol = 5
ntor = 5
stellsym = True
nfp = 3
constraint_weight = 1e0

phis = np.linspace(0, 1/nfp, 25, endpoint=False)
thetas = np.linspace(0, 1, 25, endpoint=False)
s = SurfaceRZFourier(
    mpol=mpol, ntor=ntor, stellsym=stellsym, nfp=nfp, quadpoints_phi=phis,
    quadpoints_theta=thetas)
s.fit_to_curve(ma, 0.2, flip_theta=True)

# First optimize at fixed volume

qfm = QfmResidual(s, bs)
qfm.J()

vol = Volume(s)
vol_target = vol.J()

qfm_surface = QfmSurface(bs, s, vol, vol_target)

res = qfm_surface.minimize_qfm_penalty_constraints_LBFGS(tol=1e-12, maxiter=1000,
                                                         constraint_weight=constraint_weight)
print(f"||vol constraint||={0.5*(s.volume()-vol_target)**2:.8e}, ||residual||={np.linalg.norm(qfm.J()):.8e}")

res = qfm_surface.minimize_qfm_exact_constraints_SLSQP(tol=1e-12, maxiter=1000)
print(f"||vol constraint||={0.5*(s.volume()-vol_target)**2:.8e}, ||residual||={np.linalg.norm(qfm.J()):.8e}")

# Now optimize at fixed toroidal flux

tf = ToroidalFlux(s, bs_tf)
tf_target = tf.J()

qfm_surface = QfmSurface(bs, s, tf, tf_target)

res = qfm_surface.minimize_qfm_penalty_constraints_LBFGS(tol=1e-12, maxiter=1000,
                                                         constraint_weight=constraint_weight)
print(f"||tf constraint||={0.5*(s.volume()-vol_target)**2:.8e}, ||residual||={np.linalg.norm(qfm.J()):.8e}")

res = qfm_surface.minimize_qfm_exact_constraints_SLSQP(tol=1e-12, maxiter=1000)
print(f"||tf constraint||={0.5*(tf.J()-tf_target)**2:.8e}, ||residual||={np.linalg.norm(qfm.J()):.8e}")

# Check that volume is not changed
print(f"||vol constraint||={0.5*(vol.J()-vol_target)**2:.8e}")

# Now optimize at fixed area

ar = Area(s)
ar_target = ar.J()

qfm_surface = QfmSurface(bs, s, ar, ar_target)

res = qfm_surface.minimize_qfm_penalty_constraints_LBFGS(tol=1e-12, maxiter=1000,
                                                         constraint_weight=constraint_weight)
print(f"||area constraint||={0.5*(ar.J()-ar_target)**2:.8e}, ||residual||={np.linalg.norm(qfm.J()):.8e}")

res = qfm_surface.minimize_qfm_exact_constraints_SLSQP(tol=1e-12, maxiter=1000)
print(f"||area constraint||={0.5*(ar.J()-ar_target)**2:.8e}, ||residual||={np.linalg.norm(qfm.J()):.8e}")

# Check that volume is not changed
print(f"||vol constraint||={0.5*(vol.J()-vol_target)**2:.8e}")

if "DISPLAY" in os.environ:
    s.plot()



from simsopt.field.coil import coils_via_symmetries
from simsopt.field.biotsavart import BiotSavart
from simsopt.util.zoo import get_ncsx_data
from simsopt.field.tracing import trace_particles_starting_on_curve, compute_fieldlines
from simsopt.field.magneticfieldclasses import InterpolatedField, UniformInterpolationRule
from simsopt.util.constants import PROTON_MASS, ELEMENTARY_CHARGE, ONE_EV
import simsoptpp as sopp
import numpy as np
import unittest
import logging
logging.basicConfig()
try:
    from mpi4py import MPI
    with_mpi = True
except ImportError:
    with_mpi = False


class MPITracingTesting(unittest.TestCase):

    def __init__(self, *args, **kwargs):
        super(MPITracingTesting, self).__init__(*args, **kwargs)
        logger = logging.getLogger('simsopt.field.tracing')
        logger.setLevel(1)
        curves, currents, ma = get_ncsx_data()
        nfp = 3
        coils = coils_via_symmetries(curves, currents, nfp, True)
        bs = BiotSavart(coils)
        n = 10
        rrange = (1.1, 1.8, n)
        phirange = (0, 2*np.pi/nfp, n*2)
        zrange = (0, 0.3, n//2)
        bsh = InterpolatedField(
            bs, UniformInterpolationRule(3),
            rrange, phirange, zrange, True, nfp=3, stellsym=True
        )
        bsh.estimate_error_GradAbsB(1000)
        bsh.estimate_error_B(1000)
        # trick to clear the cache in the biot savart class to reduce memory usage
        bs.set_points(np.asarray([[0., 0., 0.]])).GradAbsB()
        self.bsh = bsh
        self.ma = ma

    @unittest.skipIf(not with_mpi, "mpi not found")
    def test_parallel_fieldline(self):
        nlines = 4
        r0 = np.linalg.norm(self.ma.gamma()[0, :2])
        z0 = self.ma.gamma()[0, 2]
        R0 = [r0 + i*0.01 for i in range(nlines)]
        Z0 = [z0 for i in range(nlines)]
        phis = []
        comm = MPI.COMM_WORLD
        print('comm.size', comm.size)
        res_tys_mpi, res_phi_hits_mpi = compute_fieldlines(
            self.bsh, R0, Z0, tmax=1000, phis=phis, stopping_criteria=[], comm=comm)

        res_tys, res_phi_hits = compute_fieldlines(
            self.bsh, R0, Z0, tmax=1000, phis=phis, stopping_criteria=[], comm=None)
        for i in range(nlines):
            assert np.allclose(res_phi_hits_mpi[i], res_phi_hits[i], atol=1e-9, rtol=1e-9)

    @unittest.skipIf(not with_mpi, "mpi not found")
    def test_parallel_guiding_center(self):
        nparticles = 4
        m = PROTON_MASS
        q = ELEMENTARY_CHARGE
        Ekin = 1000*ONE_EV
        umin = 0.25
        umax = 0.75
        tmax = 1e-5
        np.set_printoptions(precision=20)
        comm = MPI.COMM_WORLD

        print('comm.size', comm.size)
        gc_tys_mpi, gc_phi_hits_mpi = trace_particles_starting_on_curve(
            self.ma, self.bsh, nparticles, tmax=tmax, seed=1, mass=m, charge=q,
            Ekin=Ekin, umin=umin, umax=umax,
            phis=[], mode='gc_vac', comm=comm)

        gc_tys, gc_phi_hits = trace_particles_starting_on_curve(
            self.ma, self.bsh, nparticles, tmax=tmax, seed=1, mass=m, charge=q,
            Ekin=Ekin, umin=umin, umax=umax,
            phis=[], mode='gc_vac', comm=None)
        for i in range(nparticles):
            assert np.allclose(gc_phi_hits_mpi[i], gc_phi_hits[i], atol=1e-9, rtol=1e-9)



#!/usr/bin/env python

from simsopt.mhd import Vmec, Boozer, Quasisymmetry
from simsopt import LeastSquaresProblem
from simsopt import least_squares_serial_solve
import os

"""
This script solve the problem in
https://github.com/landreman/stellopt_scenarios/tree/master/2DOF_circularCrossSection_varyAxis_targetIotaAndQuasisymmetry
See that website for a detailed description of the problem and plots
of the objective function landscape.
"""

vmec = Vmec(os.path.join(os.path.dirname(__file__), 'inputs', 'input.2DOF_circularCrossSection_varyAxis_targetIotaAndQuasisymmetry'))

# Define parameter space:
vmec.boundary.all_fixed()
vmec.boundary.set_fixed("rc(0,1)", False)
vmec.boundary.set_fixed("zs(0,1)", False)

# Define objective function:
boozer = Boozer(vmec, mpol=32, ntor=16)
qs = Quasisymmetry(boozer,
                   1.0,  # Radius to target
                   1, 0,  # (M, N) you want in |B|
                   normalization="symmetric",
                   weight="stellopt_ornl")

# Objective function is 100 * (iota - (-0.41))^2 + 1 * (qs - 0)^2
prob = LeastSquaresProblem([(vmec.iota_axis, -0.41, 100),
                            (qs, 0, 1)])

least_squares_serial_solve(prob)

print("Final values before shifting and scaling:", prob.dofs.f())
print("Final residuals:", prob.f())
print("Final state vector:", prob.x)
print("Final iota on axis:", vmec.iota_axis())
